from importlib.metadata import metadata
from typing import Literal
from pathlib import Path

__version__ = metadata(str(__package__))["Version"]
__author__ = metadata(str(__package__))["Author"]

from omas import ODS, load_omas_json

from device_inductance import model_reduction, contour, sensors
from device_inductance.device import DeviceInductance, TypicalOutputs
from device_inductance.coils import Coil, CoilFilament
from device_inductance.structures import PassiveStructureFilament

def load_default_ods() -> ODS:
    """Load an example ODS file in the format required by device_inductance."""
    # NOTE: This should be rewritten to use importlib once omas supports loading raw text
    ods_filename = Path(__file__).parent / "../examples/OS_SPARC_Device_Description.json"
    with open(ods_filename) as f:
        ods = load_omas_json(f)

    # Add some dummy sensors to exercise the sensor functions
    #   Full flux loop
    ffloop = ods["magnetics.flux_loop.0"]
    ffloop["type.index"] = 1
    ffloop["name"] = "dummy_full_flux_loop"
    ffloop["position.0.r"] = 3.0
    ffloop["position.0.z"] = 0.0
    #   Partial flux loop
    pfloop = ods["magnetics.flux_loop.0"]
    pfloop["type.index"] = 2
    pfloop["name"] = "dummy_partial_flux_loop"
    pfloop["position.0.r"] = 3.0
    pfloop["position.0.phi"] = 0.0
    pfloop["position.0.z"] = 0.0
    pfloop["position.1.r"] = 3.3
    pfloop["position.1.phi"] = 0.25
    pfloop["position.1.z"] = 1.0
    #   Bpol probe
    bpcoil = ods["magnetics.b_field_pol_probe.0"]
    bpcoil["type.index"] = 2
    bpcoil["name"] = "dummy_poloidal_b_field_probe"
    bpcoil["position.r"] = 3.1
    bpcoil["position.phi"] = 0.0
    bpcoil["position.z"] = 0.5
    bpcoil["poloidal_angle"] = 0.7

    return ods

def typical(
    ods: ODS,
    extent: tuple[float, float, float, float] = (0.0, 0.0, 0.0, 0.0),
    dxgrid: tuple[float, float] = (0.0, 0.0),
    max_nmodes: int = 40,
    model_reduction_method: Literal["eigenmode", "stabilized eigenmode"] = "eigenmode",
    show_prog: bool = True,
) -> TypicalOutputs:
    """
    Generate a typical set of outputs,
    notably excluding the plasma flux tables which usually require much more
    run time than the rest of the outputs combined.

    Note: during initialization, the extent of the computational grid may be
    adjusted to achieve the target spatial resolution. The adjusted extent
    will always bound the requested extent.

    Args:
        ods: An OMAS object in the format produced by device_description
        extent: [m] Extent of computational domain; adjusted during init
        dxgrid: [m] Spatial resolution of computational grid
        max_nmodes: Maximum number of structure modes to keep. Defaults to 40.
        show_prog: Whether to show terminal progress bars. Defaults to True.

    Returns:
        A fully-computed set of matrices and tables covering the needs of a typical workflow
    """
    device = DeviceInductance(
        ods=ods,
        max_nmodes=max_nmodes,
        extent=extent,
        dxgrid=dxgrid,
        model_reduction_method=model_reduction_method,
        show_prog=show_prog,
    )

    out = TypicalOutputs(
        device=device,
        extent=device.extent,
        dxgrid=device.dxgrid,
        meshes=device.meshes,
        grids=device.grids,
        extent_for_plotting=device.extent_for_plotting,
        mcc=device.coil_mutual_inductances,
        mss=device.structure_mutual_inductances,
        mcs=device.coil_structure_mutual_inductances,
        r_s=device.structure_resistances,
        r_c=device.coil_resistances,
        r_modes=device.structure_mode_resistances,
        tuv=device.structure_model_reduction,
        nmodes=device.n_structure_modes,
        psi_c=device.coil_flux_tables,
        psi_s=device.structure_flux_tables,
        psi_modes=device.structure_mode_flux_tables,
    )
    return out


__all__ = [
    "DeviceInductance",
    "TypicalOutputs",
    "typical",
    "Coil",
    "CoilFilament",
    "PassiveStructureFilament",
    "model_reduction",
    "load_default_ods",
    "contour",
    "sensors",
]



from datetime import datetime
from typing import TypeVar, Iterator

import numpy as np
from numpy.typing import NDArray


T = TypeVar("T")

# Finite-difference coeffs
# https://en.wikipedia.org/wiki/Finite_difference_coefficient
_DDX_CENTRAL_ORDER4 = np.array(
    [
        # 4th-order central difference for first derivative
        (-2, 1 / 12),
        (-1, -2 / 3),
        # (0, 0.0),
        (1, 2 / 3),
        (2, -1 / 12),
    ]
)

_DDX_FWD_ORDER4 = np.array(
    [
        # 4th-order forward difference for first derivative
        (0, -25 / 12),
        (1, 4),
        (2, -3),
        (3, 4 / 3),
        (4, -1 / 4),
    ]
)

_DDX_BWD_ORDER4 = -_DDX_FWD_ORDER4  # Reverse & flip signs


def _progressbar(it: list[T], suffix="", show_every: int = 1) -> Iterator[T]:
    """A simple terminal progressbar."""
    size = 30
    count = len(it)

    start = datetime.now()

    def show(j):
        x = int(size * j / count)
        print(
            "[{}{}] {}/{} {}, Elapsed: {:.3f} [s]".format(
                "#" * x,
                "." * (size - x),
                j,
                count,
                suffix,
                (datetime.now() - start).total_seconds(),
            ),
            end="\r",
            flush=True,
        )

    show(0)
    for i, item in enumerate(it):
        yield item
        if (i % show_every == 0) or (i == count - 1):
            show(i + 1)
    print("\n", flush=True)


def gradient_order4(
    z: NDArray, xmesh: NDArray, ymesh: NDArray
) -> tuple[NDArray, NDArray]:
    """
    Calculate gradient by 4th-order finite difference.

    `numpy.gradient` exists and is fast and convenient, but only uses a second-order difference,
    which produces unacceptable error in B-fields (well over 1% for typical geometries).

    ## Errors

        * If the input grids are not regular
        * If any input grid dimensions have size less than 5

    ## References

        * [1] “Finite difference coefficient,” Wikipedia. Aug. 22, 2023.
              Accessed: Mar. 29, 2024. [Online].
              Available: https://en.wikipedia.org/w/index.php?title=Finite_difference_coefficient

    Args:
        z: [<xunits>] 2D array of values on which to calculate the gradient
        xmesh: [m] 2D array of coordinates of first dimension
        ymesh: [m] 2D array of coordinates of second dimension

    Returns:
        (dzdx, dzdy) [<xunits>/m] 2D arrays of gradient components
    """
    nx, ny = z.shape
    dx = xmesh[1][0] - xmesh[0][0]
    dy = ymesh[0][1] - ymesh[0][0]

    # Check regular grid assumption
    assert np.all(
        np.abs(np.diff(xmesh[:, 0]) - dx) / dx < 1e-6
    ), "This method is only implemented for a regular grid"
    assert np.all(
        np.abs(np.diff(ymesh[0, :]) - dy) / dy < 1e-6
    ), "This method is only implemented for a regular grid"

    dzdx = np.zeros_like(z)
    for offs, w in _DDX_CENTRAL_ORDER4:
        start = int(2 + offs)
        end = int(nx - 2 + offs)
        dzdx[2:-2, :] += (
            w * z[start:end, :] / dx
        )  # Central difference on interior points
    for offs, w in _DDX_FWD_ORDER4:
        offs = int(offs)
        dzdx[0:2, :] += (
            w * z[offs : offs + 2, :] / dx
        )  # One-sided difference on left side
    for offs, w in _DDX_BWD_ORDER4:
        start = int(-2 + offs)
        end = int(nx + offs)
        dzdx[-2:, :] += w * z[start:end, :] / dx  # right side

    dzdy = np.zeros_like(z)
    for offs, w in _DDX_CENTRAL_ORDER4:
        start = int(2 + offs)
        end = int(ny - 2 + offs)
        dzdy[:, 2:-2] += w * z[:, start:end] / dy  # Interior points
    for offs, w in _DDX_FWD_ORDER4:
        offs = int(offs)
        dzdy[:, 0:2] += w * z[:, offs : offs + 2] / dy  # One-sided difference on bottom
    for offs, w in _DDX_BWD_ORDER4:
        start = int(-2 + offs)
        end = int(ny + offs)
        dzdy[:, -2:] += w * z[:, start:end] / dy  # top

    return dzdx, dzdy


def calc_flux_density_from_flux(
    psi: NDArray, rmesh: NDArray, zmesh: NDArray
) -> tuple[NDArray, NDArray]:
    """
    Back-calculate B-field from poloidal flux per Wesson eqn 3.2.2 by 4th-order finite difference,
    modified to use total poloidal flux instead of flux per radian.

    This avoids an expensive sum over filamentized contributions at the expense of some numerical error.

    # Errors

        * If the input grids are not regular
        * If any input grid dimensions have size less than 5

    # References

        * [1] J. Wesson, Tokamaks. Oxford, New York: Clarendon Press, 1987.

    Args:
        psi: [Wb] poloidal flux
        rmesh: [m] 2D r-coordinates
        zmesh: [m] 2D z-coordinates

    Returns:
        (br, bz) [T] 2D arrays of poloidal flux density
    """

    dpsidr, dpsidz = gradient_order4(psi, rmesh, zmesh)

    r_inv = rmesh**-1

    br = -r_inv * dpsidz / (2.0 * np.pi)  # [T]
    bz = r_inv * dpsidr / (2.0 * np.pi)  # [T]

    return (br, bz)


from sys import intern
from dataclasses import dataclass

import numpy as np

from shapely import Polygon, centroid
from omas import ODS

from cfsem import self_inductance_lyle6
from device_inductance.mesh import _mesh_region
from device_inductance.utils import _progressbar


@dataclass(frozen=True)
class PassiveStructureFilament:
    """A chunk of a cylindrically-symmetric passive conductor"""
    parent_name: str  # Name of structure that this is associated with
    r: float  # [m]
    z: float  # [m]
    area: float  # [m^2] cross-section area
    resistance: float  # [Ohm] loop resistance
    self_inductance: float  # [H]
    polygon: Polygon  # Outline of mesh element


def _extract_structures(
    description: ODS, show_prog: bool = True
) -> list[PassiveStructureFilament]:
    """
    Extract and filamentize passive conducting structural elements
    by first meshing the 2D cross-sections to quads, then collapsing
    each quad to a singularity filament at the mean position of its vertices.

    Resistance and self-inductance are estimated by approximating each quad as
    a square of equivalent area, which may give significant error if quads of
    high aspect ratio or nonconvex shape are encountered. Because the meshing
    algorithm nominally guarantees convex elements and maximum element size is
    limited in order to keep spatial resolution, there is some hope that this
    approximation will not be too far off.

    Args:
        description: Device geometric info in the format produced by device_description
        show_prog: Display a terminal progressbar

    Returns:
        A list of passive filaments, populated with reduced geometric info and estimated self-inductances.
    """
    passive_filaments: list[PassiveStructureFilament] = []

    # Use block-representation to capture different effective resistivity in different regions
    # Inner wall, outer wall, ??
    items = description["wall.description_2d.0.vessel"]["unit"].items()
    if show_prog:
        items = _progressbar(items, "Passive structure meshes")
    for partial_name, wall_section in items:
        for wall_elem in wall_section["element"].values():  # Segments of each wall
            # Unpack
            rs = wall_elem["outline.r"]
            zs = wall_elem["outline.z"]
            resistivity = wall_elem["resistivity"]  # [Ohm-m]

            # Mesh outline and approximate as filaments
            mesh = _mesh_region(np.array([x for x in zip(rs, zs)]))
            parent_name = intern(f"wall.description_2d.0.vessel.unit.{partial_name}")
            for mesh_elem in mesh:
                fil = _mesh_elem_to_fil(mesh_elem, resistivity, parent_name)
                if fil.resistance > 0.0:
                    passive_filaments.append(fil)
                else:
                    print("Skipped zero-resistance passive filament")

    # PF-related passive structures are stored separately.
    # These are vertical stability coil covers, stabilizer bars, etc
    items = description["pf_passive.loop"].items()
    if show_prog:
        items = _progressbar(items, "Passive coil meshes")
    for partial_name, passive_elem in items:
        # Unpack
        rs = passive_elem["element.0.geometry.outline.r"]
        zs = passive_elem["element.0.geometry.outline.z"]
        resistivity = passive_elem["resistivity"]

        # Mesh outline and approximate as filaments
        mesh = _mesh_region(np.array([x for x in zip(rs, zs)]))
        parent_name = intern(f"pf_passive.loop.{partial_name}")
        for mesh_elem in mesh:
            fil = _mesh_elem_to_fil(mesh_elem, resistivity, parent_name)
            if fil.resistance > 0.0:
                passive_filaments.append(fil)
            else:
                # TODO: logging instead
                print("Skipped zero-resistance passive filament")

    return passive_filaments


def _mesh_elem_to_fil(
    mesh_elem: Polygon, resistivity: float, parent_name: str
) -> PassiveStructureFilament:
    """Convert mesh element polygon to a resistive filament"""
    # Will this implicitly close? Is the mesh data ordered properly?
    area = mesh_elem.area  # [m^2]

    # Mean r,z of nodes is only really correct centroid for a proper rectangle or rhombus
    # but it should be fairly close for well-behaved shapes
    c = centroid(mesh_elem)  # [m] mean of polygon vertices
    r, z = (c.x, c.y)  # [m]
    toroidal_circumference = 2.0 * np.pi * r  # [m]
    resistance = toroidal_circumference * resistivity / area  # [Ohm]

    # Approximate as square section to estimate self-inductance
    # TODO: think more about how to handle different geometries
    # when we're ready to split hairs
    rs, zs = mesh_elem.exterior.xy
    w = max(rs) - min(rs)
    h = max(zs) - min(zs)
    self_inductance = self_inductance_lyle6(r, w, h, n=1.0)

    return PassiveStructureFilament(
        parent_name=parent_name,
        r=r,
        z=z,
        area=area,
        resistance=resistance,
        self_inductance=self_inductance,
        polygon=mesh_elem,
    )
